#pragma once

#include "HuffmanTree.h"
#include "HuffmanAbstract.h"

struct Code
{
	unsigned int code;
	unsigned int length;

	Code()
		: code(0), length(0) {}

	Code(unsigned int _code, unsigned int _length)
		: code(_code), length(_length) {}
};

typedef std::array<Code, CODEBOOKSIZE> CodeArray;



class CPUHuffman : public HuffmanCompress
{
public:
	virtual CompressedChunk compress(DataChunk& data)
	{
		CodeBookCounting cbCount = self.countDataChunk(data);
		HuffmanNode* prList = self.createTree(cbCount);
		CodeArray codeBook = self.createCodebook(prList);

		CompressedChunk compressed = self.compressHuffman(data, codeBook);
		compressed.cbFrequency = cbCount;

		return compressed;
	}
	
	virtual DataChunk decompress(CompressedChunk& data)
	{
		HuffmanNode* prList = self.createTree(data.cbFrequency);
		DataChunk result = self.decompressHuffman(data, prList);

		return result;
	}

	CodeBookCounting countDataChunk(DataChunk& data)
	{
		CodeBookCounting result;
		result.fill(0);

		auto end = data.end();
		for (auto it = data.begin(); it != end; it++)
			result[*it]++;

		// Normalize on byte values.
		unsigned int max = 0;
		for (auto it = result.begin(); it != result.end(); it++)
			max = std::max(max, *it);
		for (auto it = result.begin(); it != result.end(); it++)
			*it = (254 * *it ) / max;
			
		return result;
	}

	HuffmanNode* createTree(CodeBookCounting& counting)
	{
		NodeSet remaining;

		for (unsigned int i = 0; i < CODEBOOKSIZE; i++)
			if (counting[i] != 0)
				remaining.insert(std::pair<unsigned int, HuffmanNode*>(counting[i], new HuffmanNode(counting[i], i)));

		while (remaining.size() != 1)
		{
			auto it = remaining.begin();
			HuffmanNode* left = it->second;
			remaining.erase(it);

			it = remaining.begin();
			HuffmanNode* right = it->second;
			remaining.erase(it);

			HuffmanNode* newNode = new HuffmanNode(left, right);
			remaining.insert(std::pair<unsigned int, HuffmanNode*>(newNode->frequency, newNode));
		}
			
		auto result = remaining.begin()->second;
		if (result->left == NULL && result->right == NULL)
			result->left = new HuffmanNode(result->frequency, result->letter);

		return result;
	}


	void updateAndTraverse(HuffmanNode* tree, unsigned int currentCode, int depth, CodeArray& result)
	{
		if (tree->left != NULL)
			updateAndTraverse(tree->left, currentCode, depth+1, result);
		if (tree->right != NULL)
			updateAndTraverse(tree->right, currentCode | (1 << depth), depth+1, result);

		if (tree->left == NULL && tree->right == NULL)
			result[tree->letter] = Code(currentCode, depth);
	}

	CodeArray createCodebook(HuffmanNode* tree)
	{
		CodeArray result;
		result.fill(Code());
		updateAndTraverse(tree, 0, 0, result);

		return result;
	}

	CompressedChunk compressHuffman(DataChunk& source, CodeArray& code)
	{
		CompressedChunk result;

		unsigned int current = 0;
		unsigned int length = 0;
		for (unsigned char token : source)
		{
			Code currentCode = code[token];
			current |= currentCode.code << length;
			length += currentCode.length;
			if (length >= 8)
			{
				unsigned char pushChar = (unsigned char)(current & 255);
				result.data.push_back(pushChar);
				current >>= 8;
				length -= 8;
			}
		}
		if (length != 0)
			result.data.push_back((unsigned char)current);
		result.bitOffset = length;

		return result;
	}

	DataChunk decompressHuffman(CompressedChunk& data, HuffmanNode* tree)
	{
		DataChunk result;

		unsigned int current = 0;
		unsigned int currentLength = 0;
		unsigned int streamOffset = 0;

		unsigned int bits = data.data.size() * 8 - (8 - data.bitOffset);
		if (data.bitOffset == 0)
			bits += 8;

		HuffmanNode* currentNode = tree;

		while (bits)
		{
			if (currentLength == 0)
			{
				current = (unsigned char)data.data[streamOffset++];
				currentLength = 8;
			}

			if (current & 1)
				currentNode = currentNode->right;
			else
				currentNode = currentNode->left;
			
			current >>= 1;
			currentLength--;
			bits--;

			if (currentNode->right == NULL && currentNode->left == NULL)
			{
				result.push_back(currentNode->letter);
				currentNode = tree;
			}
		}
		return result;
	}
};