#pragma once


#include "HuffmanAbstract.h"
#include "Common.h"

#include "ConstantsCL.h"

#include <CL/opencl.h>

#undef min
#undef max

#define PARTIAL_HISTOGRAM256_COUNT 240
#define HISTOGRAM256_BIN_COUNT 256


class GPUHuffman : public HuffmanCompress
{
private:
	cl_context			context;
	cl_command_queue	commandQueue;
	cl_platform_id		platform;
	cl_device_id		device;

	cl_program			histogramProgram;
	cl_program			compressProgram;
	cl_program			scanProgram;
	cl_program			concatenationProgram;

	cl_kernel			histogramKernel;
	cl_kernel			mergeKernel;
	cl_kernel			compressKernel;
	cl_kernel			scanNaiveKernel;
	cl_kernel			scanKernel;
	cl_kernel			scanAddKernel;
	cl_kernel			concatenationKernel;

	cl_mem				subHistograms;
	cl_mem				histogram;

public:
	GPUHuffman()
		: context(NULL), commandQueue(NULL), platform(NULL), device(NULL), 
		histogramProgram(NULL), compressProgram(NULL), scanProgram(NULL), concatenationProgram(NULL),
		histogramKernel(NULL), mergeKernel(NULL), compressKernel(NULL), scanNaiveKernel(NULL), concatenationKernel(NULL),
		scanKernel(NULL), scanAddKernel(NULL),
		subHistograms(NULL), histogram(NULL)
	{
		initCL();
		initResources();
	}

	GPUHuffman(const cl_context& _context, const cl_command_queue& _commandQueue, const cl_platform_id& _platform, const cl_device_id& _device)
		: context(_context), commandQueue(_commandQueue), platform(_platform), device(_device),
		histogramProgram(NULL), compressProgram(NULL), scanProgram(NULL), concatenationProgram(NULL),
		histogramKernel(NULL), mergeKernel(NULL), compressKernel(NULL), scanNaiveKernel(NULL), concatenationKernel(NULL),
		scanKernel(NULL), scanAddKernel(NULL),
		subHistograms(NULL), histogram(NULL)

	{
		initResources();
	}

	~GPUHuffman()
	{
		releaseResources();
	}

	void releaseResources()
	{
		SAFE_RELEASE_MEMOBJECT(subHistograms);
		SAFE_RELEASE_MEMOBJECT(histogram);

		SAFE_RELEASE_KERNEL(histogramKernel);
		SAFE_RELEASE_KERNEL(mergeKernel);
		SAFE_RELEASE_KERNEL(compressKernel);
		SAFE_RELEASE_KERNEL(scanNaiveKernel);
		SAFE_RELEASE_KERNEL(concatenationKernel);

		SAFE_RELEASE_KERNEL(scanAddKernel);
		SAFE_RELEASE_KERNEL(scanKernel);

		SAFE_RELEASE_PROGRAM(histogramProgram);
		SAFE_RELEASE_PROGRAM(compressProgram);
		SAFE_RELEASE_PROGRAM(scanProgram);
		SAFE_RELEASE_PROGRAM(concatenationProgram);
	}

	virtual CompressedChunk compress(DataChunk& data)
	{
		size_t localWorksize[3] = { LOCALWORKSIZE, 1, 1 };

		// Copy source data to video memory.
		cl_int clError;
		cl_mem clData = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, data.size(), data.data(), &clError);
		V_THROW_CL(clError, "Failed to create CL data buffer");
		size_t dataSize = data.size();

		// Calculate code book.
		CodeBookCounting cbCount = self.createHistogram(clData, dataSize);

		CPUHuffman codec;
		HuffmanNode* prList = codec.createTree(cbCount);
		CodeArray codeBook = codec.createCodebook(prList);

		// Compress to chunks.
		size_t globalWorksize;
		cl_mem clCompressedChunks;
		cl_mem clChunkLengthList;
		self.compressChunks(codeBook, dataSize, clData, clCompressedChunks, clChunkLengthList, globalWorksize, localWorksize);

		// Calculate offsets for data chunk concatenation.
		// Parallel prefix sum.
		cl_mem clPrefixSum = self.scanBuffer(clChunkLengthList, globalWorksize, localWorksize);
		
		// Unfortunatetly the efficient is buggy.
		// cl_mem clPrefixSum = self.scanWorkEfficient(clChunkLengthList, globalWorksize, localWorksize);

		// Concatenate all chunks.
		CompressedChunk compressed = self.concatenate(clCompressedChunks, clChunkLengthList, clPrefixSum, globalWorksize, localWorksize);
		compressed.cbFrequency = cbCount;

		SAFE_RELEASE_MEMOBJECT(clData);
		SAFE_RELEASE_MEMOBJECT(clCompressedChunks);
		SAFE_RELEASE_MEMOBJECT(clChunkLengthList);
		SAFE_RELEASE_MEMOBJECT(clPrefixSum);

		return compressed;

	}

	virtual DataChunk decompress(CompressedChunk& data)
	{
		CPUHuffman codec;
		return codec.decompress(data);
	}

	void analyzeComponents(size_t size, size_t iterations)
	{
		DataChunk data;
		data.reserve(size);
		for (__int64 i = 0; i < size; i++)
			data.push_back(32 * i & 31);

		cl_int clError;
		cl_mem clData = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, data.size(), data.data(), &clError);
		V_THROW_CL(clError, "Failed to create CL data buffer");

		CodeBookCounting cbCount;

		std::cout << "Component profiling." << std::endl;
		std::cout << "--------------------" << std::endl;
		std::cout << std::endl;
		std::cout << "Size: " << size << std::endl;
		std::cout << "Iterations: " << iterations << std::endl;

		CTimer timer;

		timer.Start();
		for (unsigned int i = 0; i < iterations; i++)
		{
			cbCount = self.createHistogram(clData, size);
		}
		timer.Stop();
		std::cout << "Histogram   : " << timer.GetElapsedTime() / iterations << std::endl;

		CPUHuffman codec;
		HuffmanNode* prList = codec.createTree(cbCount);
		CodeArray codeBook = codec.createCodebook(prList);

		size_t globalWorksize = size;
		size_t localWorksize[3] = { 16, 0, 0 };

		timer.Start();
		for (unsigned int i = 0; i < iterations; i++)
		{
			globalWorksize = size;
			cl_mem clCompressedChunks;
			cl_mem clChunkLengthList;
			self.compressChunks(codeBook, size, clData, clCompressedChunks, clChunkLengthList, globalWorksize, localWorksize);

			SAFE_RELEASE_MEMOBJECT(clCompressedChunks);
			SAFE_RELEASE_MEMOBJECT(clChunkLengthList);
		}
		timer.Stop();
		std::cout << "Compress    : " << timer.GetElapsedTime() / iterations << std::endl;

		globalWorksize = size;
		cl_mem clCompressedChunks;
		cl_mem clChunkLengthList;
		self.compressChunks(codeBook, size, clData, clCompressedChunks, clChunkLengthList, globalWorksize, localWorksize);

		cl_mem clPrefixSum = self.scanBuffer(clChunkLengthList, globalWorksize, localWorksize);

		timer.Start();
		for (unsigned int i = 0; i < iterations; i++)
		{
			cl_mem temp = self.scanBuffer(clChunkLengthList, globalWorksize, localWorksize);
			SAFE_RELEASE_MEMOBJECT(temp);
		}
		timer.Stop();
		std::cout << "Scan        : " << timer.GetElapsedTime() / iterations << std::endl;


		timer.Start();
		for (unsigned int i = 0; i < iterations; i++)
		{
			CompressedChunk compressed = self.concatenate(clCompressedChunks, clChunkLengthList, clPrefixSum, globalWorksize, localWorksize);
		}
		timer.Stop();
		std::cout << "Concatenate : " << timer.GetElapsedTime() / iterations << std::endl;

		SAFE_RELEASE_MEMOBJECT(clCompressedChunks);
		SAFE_RELEASE_MEMOBJECT(clChunkLengthList);
	}

protected:

	void initProgram(const char* file, cl_program& program)
	{
		cl_int clError;
		char* programCode = NULL;
		size_t programSize = 0;
		LoadProgram(file, &programCode, &programSize);

		program = clCreateProgramWithSource(context, 1, (const char**)&programCode, &programSize, &clError);
		V_THROW_CL(clError, buildString("Failed to create %s from file.", file));

		clError = clBuildProgram(program, 1, &device, NULL, NULL, NULL);
		if (clError != CL_SUCCESS)
		{
			PrintBuildLog(program, device);
			throw new CLException("Build error");
		}

	}

	void initKernel(cl_program& program, cl_kernel& kernel, const char* name)
	{
		cl_int clError;
		kernel = clCreateKernel(program, name, &clError);
		V_THROW_CL(clError, buildString("Kernel %s could not be created.", name));
	}

	void initCL()
	{
		//error code
		cl_int clError;

		//get platform ID
		V_THROW_CL(clGetPlatformIDs(1, &platform, NULL), "Failed to get CL platform ID");

		//get a reference to the first available GPU device
		V_THROW_CL(clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL), "No GPU device found.");

		//Create a new OpenCL context on the selected device
		context = clCreateContext(0, 1, &device, NULL, NULL, &clError);
		V_THROW_CL(clError, "Failed to create OpenCL context.");

		//Finally, create the command queue. All the asynchronous commands to the device will be issued
		//from the CPU into this queue. This way the host program can continue the execution until some results
		//from that device are needed.
		commandQueue = clCreateCommandQueue(context, device, 0, &clError);
		V_THROW_CL(clError, "Failed to create the command queue in the context");
	}

	void initResources()
	{
		initProgram("Histogram.cl", histogramProgram);
		initKernel(histogramProgram, histogramKernel, "histogram");
		initKernel(histogramProgram, mergeKernel, "mergeHistogram");

		initProgram("Compress.cl", compressProgram);
		initKernel(compressProgram, compressKernel, "compress");

		initProgram("Scan.cl", scanProgram);
		initKernel(scanProgram, scanNaiveKernel, "scan");
		initKernel(scanProgram, scanKernel, "Scan_WorkEfficient");
		initKernel(scanProgram, scanAddKernel, "Scan_WorkEfficientAdd");

		initProgram("Concatenation.cl", concatenationProgram);
		initKernel(concatenationProgram, concatenationKernel, "concatenate");

		cl_int clError;
		subHistograms = clCreateBuffer(context, CL_MEM_READ_WRITE, PARTIAL_HISTOGRAM256_COUNT * HISTOGRAM256_BIN_COUNT * sizeof(unsigned int), 0, &clError);
		V_THROW_CL(clError, "Failed to create subhistogram array.");
		histogram = clCreateBuffer(context, CL_MEM_READ_WRITE, HISTOGRAM256_BIN_COUNT * sizeof(unsigned int), 0, &clError);
		V_THROW_CL(clError, "Failed to create subhistogram array.");
	}

	CodeBookCounting createHistogram(cl_mem& data, unsigned int dataSize)
	{
		cl_int clError;
		clError = clSetKernelArg(histogramKernel, 0, sizeof(cl_mem), &data);
		clError = clSetKernelArg(histogramKernel, 1, sizeof(unsigned int), &dataSize);
		clError = clSetKernelArg(histogramKernel, 2, sizeof(cl_mem), &subHistograms);
		V_THROW_CL(clError, "Could not set histogram parameter.");

		// Calculate sub histograms with a constant number of threads.
		size_t globalWorksize = HISTOGRAM256_THREADBLOCK_SIZE * PARTIAL_HISTOGRAM256_COUNT;
		size_t localWorksize[3] = { HISTOGRAM256_THREADBLOCK_SIZE, 1, 1 };

		clError = clEnqueueNDRangeKernel(commandQueue,
			histogramKernel,
			1,
			NULL,
			&globalWorksize,
			localWorksize,
			0,
			NULL,
			NULL);
		V_THROW_CL(clError, "Error executing histogram kernel!");

		// Now merge all subhistograms to one final histogram.
		cl_uint histCount = PARTIAL_HISTOGRAM256_COUNT;
		clError = clSetKernelArg(mergeKernel, 0, sizeof(cl_mem), &histogram);
		clError = clSetKernelArg(mergeKernel, 1, sizeof(cl_mem), &subHistograms);
		clError = clSetKernelArg(mergeKernel, 2, sizeof(cl_uint), &histCount);
		V_THROW_CL(clError, "Could not set histogram merge parameter.");

		globalWorksize = MERGE_THREADBLOCK_SIZE * HISTOGRAM256_BIN_COUNT;
		localWorksize[0] = MERGE_THREADBLOCK_SIZE;

		clError = clEnqueueNDRangeKernel(commandQueue,
			mergeKernel,
			1,
			NULL,
			&globalWorksize,
			localWorksize,
			0,
			NULL,
			NULL);
		V_THROW_CL(clError, "Error executing histogram kernel!");

		CodeBookCounting result;
		clError = clEnqueueReadBuffer(commandQueue, histogram, CL_TRUE, 0, CODEBOOKSIZE * sizeof(cl_uint), result.data(), 0, NULL, NULL);
		V_THROW_CL(clError, "Could not copy histgram from device to host.");

		// Normalize on byte values.
		unsigned int max = 0;
		for (auto it = result.begin(); it != result.end(); it++)
			max = std::max(max, *it);
		for (auto it = result.begin(); it != result.end(); it++)
			*it = (255 * *it) / max;

		return result;
	}

	cl_mem scanBuffer(cl_mem inputBuffer, size_t globalWorksize, size_t localWorksize[])
	{
		cl_int clError;
		unsigned int offset = 1;

		// Create auxilliary buffer.
		cl_mem clPingBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, globalWorksize * sizeof(cl_ulong), 0, &clError);
		V_THROW_CL(clError, "Failed to create CL ping pong buffer.");
		cl_mem clPongBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, globalWorksize * sizeof(cl_ulong), 0, &clError);
		V_THROW_CL(clError, "Failed to create CL ping pong buffer.");

		clError = clEnqueueCopyBuffer(commandQueue, inputBuffer, clPingBuffer, CL_FALSE, 0, globalWorksize * sizeof(cl_ulong), 0, NULL, NULL);
		V_THROW_CL(clError, "Failed to copy chunk offset list.");

		// Iterate log(size) times.
		while (offset < globalWorksize)
		{
			// bind scan kernel args
			clError = clSetKernelArg(scanNaiveKernel, 0, sizeof(cl_mem), (void*)&clPingBuffer);
			clError |= clSetKernelArg(scanNaiveKernel, 1, sizeof(cl_mem), (void*)&clPongBuffer);

			clError = clSetKernelArg(scanNaiveKernel, 2, sizeof(cl_uint), (void*)&globalWorksize);
			clError |= clSetKernelArg(scanNaiveKernel, 3, sizeof(cl_uint), (void*)&offset);
			V_THROW_CL(clError, "Failed to bind scan kernel arguments");

			// run scan kernel
			clError = clEnqueueNDRangeKernel(commandQueue, scanNaiveKernel, 1, NULL, &globalWorksize, localWorksize, 0, NULL, NULL);
			V_THROW_CL(clError, "Error executing scan kernel!");

			// Swap aux and chunklist buffer for next iteration.
			std::swap(clPingBuffer, clPongBuffer);

			offset <<= 1;
		}

		SAFE_RELEASE_MEMOBJECT(clPongBuffer);

		return clPingBuffer;
	}

	void compressChunks(CodeArray& codeBook, size_t dataSize, cl_mem clData, cl_mem& clCompressedChunks, cl_mem& clChunkLengthList, size_t& globalWorksize, size_t localWorksize[])
	{
		cl_int clError;

		// Set up work sizes for compression kernel.
		globalWorksize = 2*localWorksize[0];
		while (globalWorksize < dataSize / (DWORDPERTHREAD * 4))
			globalWorksize *= 2;
		if (globalWorksize < 256)
			globalWorksize = 256;

		// Set up buffers and parameters for compression kernel.
		cl_mem clCodeBook = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, sizeof(Code)* CODEBOOKSIZE, &codeBook, &clError);
		V_THROW_CL(clError, "Failed to create CL codebook.");
		clCompressedChunks = clCreateBuffer(context, CL_MEM_READ_WRITE, dataSize * 2, 0, &clError);
		V_THROW_CL(clError, "Failed to create CL compressed chunks.");
		clChunkLengthList = clCreateBuffer(context, CL_MEM_READ_WRITE, globalWorksize * sizeof(cl_ulong), 0, &clError);
		V_THROW_CL(clError, "Failed to create CL chunk length list.");

		clError = clSetKernelArg(compressKernel, 0, sizeof(cl_mem), &clData);
		clError |= clSetKernelArg(compressKernel, 1, sizeof(cl_uint), &(cl_uint)dataSize);
		clError |= clSetKernelArg(compressKernel, 2, sizeof(cl_mem), &clCodeBook);
		clError |= clSetKernelArg(compressKernel, 3, sizeof(cl_mem), &clCompressedChunks);
		clError |= clSetKernelArg(compressKernel, 4, sizeof(cl_mem), &clChunkLengthList);
		//clError |= clSetKernelArg(compressKernel, 5, sizeof(cl_uint) * CODEBOOKSIZE, NULL);
		//clError |= clSetKernelArg(compressKernel, 6, sizeof(cl_uint) * CODEBOOKSIZE, NULL);
		V_THROW_CL(clError, "Failed to set compress kernel parameter.");

		// Compression kernel. Creates compressed chunks.
		clError = clEnqueueNDRangeKernel(commandQueue,
			compressKernel,
			1,
			NULL,
			&globalWorksize,
			localWorksize,
			0,
			NULL,
			NULL);
		V_THROW_CL(clError, "Error executing compress kernel!");

		SAFE_RELEASE_MEMOBJECT(clCodeBook);
	}

	CompressedChunk concatenate(cl_mem clCompressedChunks, cl_mem clChunkLengthList, cl_mem clPrefixSum, size_t globalWorksize, size_t localWorksize[])
	{
		CompressedChunk result;
		cl_int clError;

		// Calculate amount of chunks and read the compressed chunk size. (Could be highly optimized!)
		cl_ulong* prefixSum = (cl_ulong*)malloc(globalWorksize * sizeof(cl_ulong));
		clError = clEnqueueReadBuffer(commandQueue, clPrefixSum, CL_TRUE, 0, globalWorksize * sizeof(cl_ulong), prefixSum, 0, NULL, NULL);
		V_THROW_CL(clError, "Could not copy prefix data from device to host.");
		cl_ulong compressedSize = prefixSum[globalWorksize - 1];
		free(prefixSum);

		cl_mem clResult = clCreateBuffer(context, CL_MEM_READ_WRITE, (size_t)(compressedSize / 8 + 1), 0, &clError);
		V_THROW_CL(clError, "Failed to create CL compressed chunks.");

		clError = clSetKernelArg(concatenationKernel, 0, sizeof(cl_mem), &clCompressedChunks);
		clError |= clSetKernelArg(concatenationKernel, 1, sizeof(cl_mem), &clResult);
		clError |= clSetKernelArg(concatenationKernel, 2, sizeof(cl_mem), &clChunkLengthList);
		clError |= clSetKernelArg(concatenationKernel, 3, sizeof(cl_mem), &clPrefixSum);
		V_THROW_CL(clError, "Failed to set concatenation kernel parameter.");

		// Compression kernel. Creates compressed chunks.
		clError = clEnqueueNDRangeKernel(commandQueue,
			concatenationKernel,
			1,
			NULL,
			&globalWorksize,
			localWorksize,
			0,
			NULL,
			NULL);
		V_THROW_CL(clError, "Error executing concatenation kernel!");


		char* temp = (char*)malloc((size_t)(compressedSize/8+1));
		clError = clEnqueueReadBuffer(commandQueue, clResult, CL_TRUE, 0, (size_t)(compressedSize/8+1), temp, 0, NULL, NULL);
		V_THROW_CL(clError, "Could not copy compressed data from device to host.");

		if (compressedSize % 8 != 0)
		{
			result.bitOffset = compressedSize % 8;
			compressedSize += 8;
		}
		else
			result.bitOffset = 0;

		result.data = DataChunk(temp, temp + compressedSize / 8);

		SAFE_RELEASE_MEMOBJECT(clResult);
		return result;
	}


	cl_mem scanWorkEfficient(cl_mem inputBuffer, size_t globalWorksize, size_t localWorksize[])
	{

		cl_int clError;

		// Calculate levels.
		unsigned int levels = 1;
		size_t size = globalWorksize/2;
		while (size > 0)
		{
			size /= 2 * localWorksize[0];
			levels++;
		}

		cl_mem* clLevelArray = new cl_mem[levels];
		unsigned int N = globalWorksize;
		for (unsigned int i = 0; i < levels; i++) 
		{
			clLevelArray[i] = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_ulong)* N, NULL, &clError);
			V_THROW_CL(clError, "Failed to initialize scan array.");
			N = std::max(N / (2 * localWorksize[0]), localWorksize[0]);
		}

		clError = clEnqueueCopyBuffer(commandQueue, inputBuffer, clLevelArray[0], CL_FALSE, 0, globalWorksize * sizeof(cl_ulong), 0, NULL, NULL);
		V_THROW_CL(clError, "Failed to copy chunk offset list.");

		globalWorksize /= 2;
		size_t local[3] = { localWorksize[0], 1, 1 };

		unsigned int minSize = 2*localWorksize[0];
		unsigned int arraySizes[1024];

		// Computing all Array blocks.
		for (unsigned int i = 0; i < levels - 1; i++)
		{
			arraySizes[i] = globalWorksize;
			local[0] = std::min(localWorksize[0], globalWorksize);
			std::cout << "Global: " << globalWorksize << " local: " << local[0] << std::endl;

			clError  = clSetKernelArg(scanKernel, 0, sizeof(cl_mem), (void *)&clLevelArray[i]);
			clError |= clSetKernelArg(scanKernel, 1, sizeof(cl_mem), (void *)&clLevelArray[i + 1]);
			clError |= clSetKernelArg(scanKernel, 2, sizeof(cl_ulong)* localWorksize[0] * 2, NULL);
			V_THROW_CL(clError, "Failed to bin scan kernel arguments.");

			//size_t global = std::max(globalWorksize, minSize);

			clError = clEnqueueNDRangeKernel(	commandQueue,
												scanKernel,
												1,
												NULL,
												&globalWorksize,
												local,
												0,
												NULL,
												NULL);
			V_THROW_CL(clError, "Error executing scan kernel!");
			globalWorksize /= minSize;
		}

		for (int i = levels - 1; i > 0; i--)
		{
			globalWorksize = arraySizes[i - 1];
			local[0] = std::min(localWorksize[0], globalWorksize);
			clError = clSetKernelArg(scanAddKernel, 0, sizeof(cl_mem), (void *)&clLevelArray[i]);
			clError |= clSetKernelArg(scanAddKernel, 1, sizeof(cl_mem), (void *)&clLevelArray[i - 1]);
			
			size_t localMemSize = localWorksize[0];
			clError |= clSetKernelArg(scanAddKernel, 2, sizeof(cl_ulong) * localMemSize, NULL);
			V_THROW_CL(clError, "Failed to set scan add parameter.");

			clError = clEnqueueNDRangeKernel(	commandQueue,
												scanAddKernel,
												1,
												NULL,
												&globalWorksize,
												local,
												0,
												NULL,
												NULL);
			V_THROW_CL(clError, "Error executing scan sum kernel!");
		}

		cl_mem result = clLevelArray[0];

		for (unsigned int i = 1; i < levels; i++)
			SAFE_RELEASE_MEMOBJECT(clLevelArray[i]);

		delete[] clLevelArray;

		return result;
	}


};


