#pragma once


#include "HuffmanAbstract.h"
#include "Common.h"

#include "ConstantsCL.h"

#include <CL/opencl.h

#define PARTIAL_HISTOGRAM256_COUNT 20
#define HISTOGRAM256_BIN_COUNT 256


class GPUHuffman : public HuffmanCompress
{
private:
	cl_context			context;
	cl_command_queue	commandQueue;
	cl_platform_id		platform;
	cl_device_id		device;

	cl_program			histogramProgram;
	cl_program			compressProgram;
	cl_program			scanProgram;
	cl_program			concatenationProgram;

	cl_kernel			histogramKernel;
	cl_kernel			mergeKernel;
	cl_kernel			compressKernel;
	cl_kernel			scanKernel;
	cl_kernel			concatenationKernel;

	cl_mem				subHistograms;
	cl_mem				histogram;
	cl_mem				clCodebook;

public:
	GPUHuffman()
		: context(NULL), commandQueue(NULL), platform(NULL), device(NULL)
	{
		initCL();
		initResources();
	}

	GPUHuffman(const cl_context& _context, const cl_command_queue& _commandQueue, const cl_platform_id& _platform, const cl_device_id& _device)
		: context(_context), commandQueue(_commandQueue), platform(_platform), device(_device)
	{
		initResources();
	}

	virtual CompressedChunk compress(DataChunk& data)
	{
		size_t localWorksize[3] = { 128, 1, 1 };

		// Copy source data to video memory.
		cl_int clError;
		cl_mem clData = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, data.size(), data.data(), &clError);
		V_THROW_CL(clError, "Failed to create CL data buffer");
		size_t dataSize = data.size();

		// Calculate code book.
		CodeBookCounting cbCount = self.createHistogram(clData, dataSize);

		CPUHuffman codec;
		//cbCount = codec.countDataChunk(data);
		HuffmanNode* prList = codec.createTree(cbCount);
		CodeArray codeBook = codec.createCodebook(prList);

		// Compress to chunks.
		size_t globalWorksize;
		cl_mem clCompressedChunks;
		cl_mem clChunkLengthList;
		self.compressChunks(codeBook, dataSize, clData, clCompressedChunks, clChunkLengthList, globalWorksize, localWorksize);

		// Calculate offsets for data chunk concatenation.
		// Parallel prefix sum.
		cl_mem clPrefixSum = self.scanBuffer(clChunkLengthList, globalWorksize, localWorksize);

		// Concatenate all chunks.
		CompressedChunk compressed = self.concatenate(clCompressedChunks, clChunkLengthList, clPrefixSum, globalWorksize, localWorksize);
		compressed.cbFrequency = cbCount;

/*		// DEBUG
		char* debugCompressed = (char*)malloc(dataSize * 2);
		cl_ulong* debugChunkLength = (cl_ulong*)malloc(globalWorksize * sizeof(cl_ulong));
		clError = clEnqueueReadBuffer(commandQueue, clCompressedChunks, CL_TRUE, 0, dataSize*2, debugCompressed, 0, NULL, NULL);
		V_THROW_CL(clError, "Could not copy compressed data from device to host.");
		clError = clEnqueueReadBuffer(commandQueue, clChunkLengthList, CL_TRUE, 0, globalWorksize * sizeof(cl_ulong), debugChunkLength, 0, NULL, NULL);
		V_THROW_CL(clError, "Could not copy chunk length list from device to host.");


		std::vector<__int6 dbgData(debugChunkLength, debugChunkLength + globalWorksize);

		//CompressedChunk compressed = codec.compressHuffman(data, codeBook);
		CompressedChunk compressed;
		debugCompressed += 2*BYTESPERTHREAD;
		compressed.data = DataChunk(debugCompressed, debugCompressed + (1+dbgData[1] / 8));
		compressed.bitOffset = dbgData[1] % 8;
		compressed.cbFrequency = cbCount;
	*/	
		return compressed;

	}

	virtual DataChunk decompress(CompressedChunk& data)
	{
		CPUHuffman codec;
		return codec.decompress(data);
	}

protected:

	void initProgram(const char* file, cl_program& program)
	{
		cl_int clError;
		char* programCode = NULL;
		size_t programSize = 0;
		LoadProgram(file, &programCode, &programSize);

		program = clCreateProgramWithSource(context, 1, (const char**)&programCode, &programSize, &clError);
		V_THROW_CL(clError, buildString("Failed to create %s from file.", file));

		clError = clBuildProgram(program, 1, &device, NULL, NULL, NULL);
		if (clError != CL_SUCCESS)
		{
			PrintBuildLog(program, device);
			throw new CLException("Build error");
		}

	}

	void initKernel(cl_program& program, cl_kernel& kernel, const char* name)
	{
		cl_int clError;
		kernel = clCreateKernel(program, name, &clError);
		V_THROW_CL(clError, buildString("Kernel %s could not be created.", name));
	}

	void initCL()
	{
		//error code
		cl_int clError;

		//get platform ID
		V_THROW_CL(clGetPlatformIDs(1, &platform, NULL), "Failed to get CL platform ID");

		//get a reference to the first available GPU device
		V_THROW_CL(clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL), "No GPU device found.");

		//Create a new OpenCL context on the selected device
		context = clCreateContext(0, 1, &device, NULL, NULL, &clError);
		V_THROW_CL(clError, "Failed to create OpenCL context.");

		//Finally, create the command queue. All the asynchronous commands to the device will be issued
		//from the CPU into this queue. This way the host program can continue the execution until some results
		//from that device are needed.
		commandQueue = clCreateCommandQueue(context, device, 0, &clError);
		V_THROW_CL(clError, "Failed to create the command queue in the context");
	}

	void initResources()
	{
		initProgram("Histogram.cl", histogramProgram);
		initKernel(histogramProgram, histogramKernel, "histogram");
		initKernel(histogramProgram, mergeKernel, "mergeHistogram");

		initProgram("Compress.cl", compressProgram);
		initKernel(compressProgram, compressKernel, "compress");

		initProgram("Scan.cl", scanProgram);
		initKernel(scanProgram, scanKernel, "scan");

		initProgram("Concatenation.cl", concatenationProgram);
		initKernel(concatenationProgram, concatenationKernel, "concatenate");

		cl_int clError;
		subHistograms = clCreateBuffer(context, CL_MEM_READ_WRITE, PARTIAL_HISTOGRAM256_COUNT * HISTOGRAM256_BIN_COUNT * sizeof(unsigned int), 0, &clError);
		V_THROW_CL(clError, "Failed to create subhistogram array.");
		histogram = clCreateBuffer(context, CL_MEM_READ_WRITE, HISTOGRAM256_BIN_COUNT * sizeof(unsigned int), 0, &clError);
		V_THROW_CL(clError, "Failed to create subhistogram array.");
	}

	CodeBookCounting createHistogram(cl_mem& data, unsigned int dataSize)
	{
		cl_int clError;
		clError = clSetKernelArg(histogramKernel, 0, sizeof(cl_mem), &data);
		clError = clSetKernelArg(histogramKernel, 1, sizeof(unsigned int), &dataSize);
		clError = clSetKernelArg(histogramKernel, 2, sizeof(cl_mem), &subHistograms);
		V_THROW_CL(clError, "Could not set histogram parameter.");

		// Calculate sub histograms with a constant number of threads.
		size_t globalWorksize = HISTOGRAM256_THREADBLOCK_SIZE * PARTIAL_HISTOGRAM256_COUNT;
		size_t localWorksize[3] = { HISTOGRAM256_THREADBLOCK_SIZE, 1, 1 };

		clError = clEnqueueNDRangeKernel(commandQueue,
			histogramKernel,
			1,
			NULL,
			&globalWorksize,
			localWorksize,
			0,
			NULL,
			NULL);
		V_THROW_CL(clError, "Error executing histogram kernel!");

		// Now merge all subhistograms to one final histogram.
		cl_uint histCount = PARTIAL_HISTOGRAM256_COUNT;
		clError = clSetKernelArg(mergeKernel, 0, sizeof(cl_mem), &histogram);
		clError = clSetKernelArg(mergeKernel, 1, sizeof(cl_mem), &subHistograms);
		clError = clSetKernelArg(mergeKernel, 2, sizeof(cl_uint), &histCount);
		V_THROW_CL(clError, "Could not set histogram merge parameter.");

		globalWorksize = MERGE_THREADBLOCK_SIZE * HISTOGRAM256_BIN_COUNT;
		localWorksize[0] = MERGE_THREADBLOCK_SIZE;

		clError = clEnqueueNDRangeKernel(commandQueue,
			mergeKernel,
			1,
			NULL,
			&globalWorksize,
			localWorksize,
			0,
			NULL,
			NULL);
		V_THROW_CL(clError, "Error executing histogram kernel!");

		CodeBookCounting result;
		clError = clEnqueueReadBuffer(commandQueue, histogram, CL_TRUE, 0, CODEBOOKSIZE * sizeof(cl_uint), result.data(), 0, NULL, NULL);
		V_THROW_CL(clError, "Could not copy histgram from device to host.");

		// Normalize on byte values.
		unsigned int max = 0;
		for (auto it = result.begin(); it != result.end(); it++)
			max = std::max(max, *it);
		for (auto it = result.begin(); it != result.end(); it++)
			*it = (255 * *it) / max;

		return result;
	}

	cl_mem scanBuffer(cl_mem inputBuffer, size_t globalWorksize, size_t localWorksize[])
	{
		cl_int clError;
		unsigned int offset = 1;

		// Create auxilliary buffer.
		cl_mem clPingBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, globalWorksize * sizeof(cl_ulong), 0, &clError);
		V_THROW_CL(clError, "Failed to create CL ping pong buffer.");
		cl_mem clPongBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, globalWorksize * sizeof(cl_ulong), 0, &clError);
		V_THROW_CL(clError, "Failed to create CL ping pong buffer.");

		clError = clEnqueueCopyBuffer(commandQueue, inputBuffer, clPingBuffer, CL_FALSE, 0, globalWorksize * sizeof(cl_ulong), 0, NULL, NULL);
		V_THROW_CL(clError, "Failed to copy chunk offset list.");

		// Iterate log(size) times.
		while (offset < globalWorksize)
		{
			// bind scan kernel args
			clError = clSetKernelArg(scanKernel, 0, sizeof(cl_mem), (void*)&clPingBuffer);
			clError |= clSetKernelArg(scanKernel, 1, sizeof(cl_mem), (void*)&clPongBuffer);

			clError = clSetKernelArg(scanKernel, 2, sizeof(cl_uint), (void*)&globalWorksize);
			clError |= clSetKernelArg(scanKernel, 3, sizeof(cl_uint), (void*)&offset);
			V_THROW_CL(clError, "Failed to bind scan kernel arguments");

			// run scan kernel
			clError = clEnqueueNDRangeKernel(commandQueue, scanKernel, 1, NULL, &globalWorksize, localWorksize, 0, NULL, NULL);
			V_THROW_CL(clError, "Error executing scan kernel!");

			// Swap aux and chunklist buffer for next iteration.
			std::swap(clPingBuffer, clPongBuffer);

			offset <<= 1;
		}

		return clPingBuffer;
	}

	void compressChunks(CodeArray& codeBook, size_t dataSize, cl_mem clData, cl_mem& clCompressedChunks, cl_mem& clChunkLengthList, size_t& globalWorksize, size_t localWorksize[])
	{
		cl_int clError;

		// Set up work sizes for compression kernel.
		globalWorksize = GetGlobalWorkSize(dataSize / (DWORDPERTHREAD * ), localWorksize[0]);
		if (globalWorksize < 256)
			globalWorksize = 256;

		// Set up buffers and parameters for compression kernel.
		cl_mem clCodeBook = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, sizeof(Code)* CODEBOOKSIZE, &codeBook, &clError);
		V_THROW_CL(clError, "Failed to create CL codebook.");
		clCompressedChunks = clCreateBuffer(context, CL_MEM_READ_WRITE, dataSize * 2, 0, &clError);
		V_THROW_CL(clError, "Failed to create CL compressed chunks.");
		clChunkLengthList = clCreateBuffer(context, CL_MEM_READ_WRITE, globalWorksize * sizeof(cl_ulong), 0, &clError);
		V_THROW_CL(clError, "Failed to create CL chunk length list.");

		clError = clSetKernelArg(compressKernel, 0, sizeof(cl_mem), &clData);
		clError |= clSetKernelArg(compressKernel, 1, sizeof(cl_uint), &(cl_uint)dataSize);
		clError |= clSetKernelArg(compressKernel, 2, sizeof(cl_mem), &clCodeBook);
		clError |= clSetKernelArg(compressKernel, 3, sizeof(cl_mem), &clCompressedChunks);
		clError |= clSetKernelArg(compressKernel, , sizeof(cl_mem), &clChunkLengthList);
		//clError |= clSetKernelArg(compressKernel, 5, sizeof(cl_uint) * CODEBOOKSIZE, NULL);
		//clError |= clSetKernelArg(compressKernel, 6, sizeof(cl_uint) * CODEBOOKSIZE, NULL);
		V_THROW_CL(clError, "Failed to set compress kernel parameter.");

		// Compression kernel. Creates compressed chunks.
		clError = clEnqueueNDRangeKernel(commandQueue,
			compressKernel,
			1,
			NULL,
			&globalWorksize,
			localWorksize,
			0,
			NULL,
			NULL);
		V_THROW_CL(clError, "Error executing compress kernel!");
	}

	CompressedChunk concatenate(cl_mem clCompressedChunks, cl_mem clChunkLengthList, cl_mem clPrefixSum, size_t globalWorksize, size_t localWorksize[])
	{
		CompressedChunk result;
		cl_int clError;

		// Calculate amount of chunks and read the compressed chunk size. (Could be highly optimized!)
		cl_ulong* prefixSum = (cl_ulong*)malloc(globalWorksize * sizeof(cl_ulong));
		clError = clEnqueueReadBuffer(commandQueue, clPrefixSum, CL_TRUE, 0, globalWorksize * sizeof(cl_ulong), prefixSum, 0, NULL, NULL);
		V_THROW_CL(clError, "Could not copy prefix data from device to host.");
		cl_ulong compressedSize = prefixSum[globalWorksize  1];
		free(prefixSum);

		cl_mem clResult = clCreateBuffer(context, CL_MEM_READ_WRITE, (size_t)(compressedSize / 8 + 1), 0, &clError);
		V_THROW_CL(clError, "Failed to create CL compressed chunks.");

		clError = clSetKernelArg(concatenationKernel, 0, sizeof(cl_mem), &clCompressedChunks);
		clError |= clSetKernelArg(concatenationKernel, 1, sizeof(cl_mem), &clResult);
		clError |= clSetKernelArg(concatenationKernel, 2, sizeof(cl_mem), &clChunkLengthList);
		clError |= clSetKernelArg(concatenationKernel, 3, sizeof(cl_mem), &clPrefixSum);
		V_THROW_CL(clError, "Failed to set concatenation kernel parameter.");

		// Compression kernel. Creates compressed chunks.
		clError = clEnqueueNDRangeKernel(commandQueue,
			concatenationKernel,
			1,
			NULL,
			&globalWorksize,
			localWorksize,
			0,
			NULL,
			NULL);
		V_THROW_CL(clError, "Error executing concatenation kernel!");


		char* temp = (char*)malloc((size_t)(compressedSize/8+1));
		clError = clEnqueueReadBuffer(commandQueue, clResult, CL_TRUE, 0, (size_t)(compressedSize/8+1), temp, 0, NULL, NULL);
		V_THROW_CL(clError, "Could not copy compressed data from device to host.");

		if (compressedSize % 8 != 0)
		{
			result.bitOffset = compressedSize % 8;
			compressedSize += 8;
		}
		else
			result.bitOffset = 0;

		result.data = DataChunk(temp, temp + compressedSize / 8);

		return result;
	}
};


