// Adapted from NVIDIA HISTOGRAM 256 CUDA EXAMPLE!

#include "ConstantsCL.h"

inline void addByte(__local volatile uint* s_WarpHist, uint data, uint threadTag)
{
    uint count;

    do
    {
        count = s_WarpHist[data] & TAG_MASK;
        count = threadTag | (count + 1);
        s_WarpHist[data] = count;
    }
    while (s_WarpHist[data] != count);
}

inline void addDWord(__local uint *s_WarpHist, uint data, uint tag)
{
    addByte(s_WarpHist, (data >>  0) & 0xFFU, tag);
    addByte(s_WarpHist, (data >>  8) & 0xFFU, tag);
    addByte(s_WarpHist, (data >> 16) & 0xFFU, tag);
    addByte(s_WarpHist, (data >> 24) & 0xFFU, tag);
}

inline void histogram256(__global uint *d_PartialHistograms, __global uint *d_Data, uint dataCount)
{
    //Per-warp subhistogram storage
    __local uint s_Hist[HISTOGRAM256_THREADBLOCK_MEMORY];
    __local uint *s_WarpHist = s_Hist + (LIDX >> LOG2_WARP_SIZE) * HISTOGRAM256_BIN_COUNT;

    //Clear shared memory storage for current threadblock before processing
	#pragma unroll
    for (uint i = 0; i < (HISTOGRAM256_THREADBLOCK_MEMORY / HISTOGRAM256_THREADBLOCK_SIZE); i++)
    {
        s_Hist[LIDX + i * HISTOGRAM256_THREADBLOCK_SIZE] = 0;
    }

    //Cycle through the entire data set, update subhistograms for each warp
    const uint tag = LIDX << (UINT_BITS - LOG2_WARP_SIZE);

	// Sync threads
	barrier(CLK_LOCAL_MEM_FENCE);

    for (uint pos = GIDX; pos < dataCount; pos += LOCALSIZEX * get_num_groups(0))
    {
        uint data = d_Data[pos];
        addDWord(s_WarpHist, data, tag);
    }

    //Merge per-warp histograms into per-block and write to global memory
	barrier(CLK_LOCAL_MEM_FENCE);

    for (uint bin = LIDX; bin < HISTOGRAM256_BIN_COUNT; bin += HISTOGRAM256_THREADBLOCK_SIZE)
    {
        uint sum = 0;

        for (uint i = 0; i < WARP_COUNT; i++)
        {
            sum += s_Hist[bin + i * HISTOGRAM256_BIN_COUNT] & TAG_MASK;
        }

        d_PartialHistograms[(GIDX / LOCALSIZEX) * HISTOGRAM256_BIN_COUNT + bin] = sum;
    }
}

__kernel void histogram(__global uchar* data,
						uint size,
						__global uint* subHistograms)
{
    // Size of data block must be multiple of 4.
	histogram256(subHistograms, (__global uint*)data, size / sizeof(uint));
	
	// manually count the missing
	uint missing = size % sizeof(uint);
	if (GIDX == 0 && missing != 0)
	{
		for (int i = 0; i < missing; i++)
			subHistograms[data[size-i-1]]++;
	}
}


__kernel void mergeHistogram(
    __global uint *d_Histogram,
    __global uint *d_PartialHistograms,
    uint histogramCount
)
{
    uint sum = 0;

    for (uint i = LIDX; i < histogramCount; i += MERGE_THREADBLOCK_SIZE)
    {
        sum += d_PartialHistograms[(GIDX / LOCALSIZEX) + i * HISTOGRAM256_BIN_COUNT];
    }

    __local uint data[MERGE_THREADBLOCK_SIZE];
    data[LIDX] = sum;

    for (uint stride = MERGE_THREADBLOCK_SIZE / 2; stride > 0; stride >>= 1)
    {
		barrier(CLK_LOCAL_MEM_FENCE);

        if (LIDX < stride)
        {
            data[LIDX] += data[LIDX + stride];
        }
    }

    if (LIDX == 0)
    {
        d_Histogram[(GIDX / LOCALSIZEX)] = data[0];
    }
}


