#pragma once

#include "Exception.h"
#include "CTimer.h"
#undef max

#include <fstream>
#include <vector>
#include <array>
#include <algorithm>

#define CODEBOOKSIZE 256

#define self (*this)

typedef std::vector<unsigned char> DataChunk;
typedef std::array<unsigned int, CODEBOOKSIZE> CodeBookCounting;

struct CompressedChunk
{
	DataChunk data;
	unsigned int bitOffset;
	CodeBookCounting cbFrequency;
};

#define CHECKFILE(__STREAM, __NAME)\
	if (__STREAM.fail())\
	throw new InvalidFileException(__NAME);

template<class STREAM> inline std::streamoff streamRemain(STREAM& stream)
{
	std::streamoff currentPos = stream.tellg();
	stream.seekg(0, stream.end);
	std::streamoff fileSize = stream.tellg() - currentPos;
	stream.seekg(currentPos, stream.beg);

	return fileSize;
}

template<class STREAM> inline std::streamoff streamSize(STREAM& stream)
{
	std::streamoff currentPos = stream.tellg();
	stream.seekg(0, stream.end);
	std::streamoff fileSize = stream.tellg();
	stream.seekg(currentPos, stream.beg);

	return fileSize;
}

std::ofstream& operator<<(std::ofstream& stream, const CompressedChunk& compressed)
{
	auto end = compressed.cbFrequency.end();
	for (auto it = compressed.cbFrequency.begin(); it != end; it++)
	{
		unsigned char dictByte = *it;
		stream.write((char*)&dictByte, 1);
	}
	unsigned char bitOffs = (unsigned char)compressed.bitOffset;
	stream.write((char*)&bitOffs, 1);

	std::streamoff dataSize = compressed.data.size();
	stream.write((char*)compressed.data.data(), dataSize);

	return stream;
}

std::ifstream& operator<<(CompressedChunk& compressed, std::ifstream& stream)
{
	unsigned char byteBuf[256];
	stream.read((char*)byteBuf, 256);
	for (int i = 0; i < 256; i++)
		compressed.cbFrequency[i] = byteBuf[i];

	char bitOffs;
	stream.read((char*)&bitOffs, 1);
	compressed.bitOffset = bitOffs;

	std::streamoff fileSize = streamRemain(stream);

	compressed.data.resize((unsigned int)(fileSize));
	stream.read((char*)compressed.data.data(), fileSize);

	return stream;
}


class HuffmanCompress
{
public:
	virtual CompressedChunk compress(DataChunk& data) = 0;
	virtual DataChunk decompress(CompressedChunk& data) = 0;

	void compress(std::string filename)
	{
		std::ifstream inputFile(filename, std::ios::binary);
		CHECKFILE(inputFile, filename);

		std::streamoff fileSize = streamRemain(inputFile);
		if (fileSize == 0)
			throw new InvalidFileException(filename);

		DataChunk chunk((unsigned int)fileSize);
		inputFile.read((char*)chunk.data(), fileSize);

		CompressedChunk compressed = compress(chunk);

		filename += ".huf";
		std::ofstream outputFile(filename, std::ios::binary);
		CHECKFILE(outputFile, filename);
		outputFile.clear();

		outputFile << compressed;
	}

	void decompress(std::string filename, std::string newName)
	{
		std::ifstream inputFile(filename, std::ios::binary);
		if (inputFile.fail())
			throw new InvalidFileException(filename);

		CompressedChunk compressed;
		compressed << inputFile;

		DataChunk decompressed = decompress(compressed);

		std::ofstream outputFile(newName, std::ios::binary);
		if (outputFile.fail())
			throw new InvalidFileException(newName);
		outputFile.clear();

		outputFile.write((char*)&decompressed[0], decompressed.size());
	}

	void decompress(std::string filename)
	{
		std::string newName(filename);
		newName.erase(filename.find_last_of("."), std::string::npos);
		decompress(filename, newName);
	}

	double stressTest(DataChunk& data)
	{
		CTimer timer;
		unsigned int iterations = 100000 / (unsigned int)floor(sqrt(data.size()));
		iterations = std::max(iterations, (unsigned int)4);
		iterations = 1;

		timer.Start();
		for (unsigned int i = 0; i < iterations; i++)
			CompressedChunk result = self.compress(data);
		timer.Stop();

		return timer.GetElapsedTime() / iterations;
	}

};

